from typing import NoReturn


def never(x: NoReturn, /) -> NoReturn:
    """Never return. Used for exhaustive pattern matching."""

    raise Never(f'"never" was run with {x}')


class Never(Exception):
    """Raised when `never` is run."""



from pathlib import Path

from beartype import beartype
from hypothesis.strategies import SearchStrategy

from utilities.hypothesis import draw_and_map
from utilities.hypothesis import temp_paths


@beartype
def task_namespaces() -> SearchStrategy[str]:
    """Strategy for generating task namespaces."""

    return draw_and_map(_draw_task_namespace, temp_paths())


@beartype
def _draw_task_namespace(path: Path, /) -> str:
    return path.name



from typing import NoReturn

from pytest import raises

from utilities.errors import NoUniqueArg
from utilities.errors import redirect_error


class TestRedirectError:
    def test_redirected(self) -> None:
        def raises_generic() -> NoReturn:
            raise ValueError("error")

        class CustomError(ValueError):
            ...

        with raises(CustomError):
            try:
                raises_generic()
            except ValueError as error:
                redirect_error(error, "error", CustomError)

    def test_not_redirected(self) -> None:
        def raises_generic() -> NoReturn:
            raise ValueError("error")

        class CustomError(ValueError):
            ...

        with raises(ValueError):
            try:
                raises_generic()
            except ValueError as error:
                redirect_error(error, "something else", CustomError)

    def test_no_unique_arg(self) -> None:
        def raises_generic() -> NoReturn:
            raise ValueError(0, 1)

        with raises(NoUniqueArg):
            try:
                raises_generic()
            except ValueError as error:
                redirect_error(error, "error", RuntimeError)



from pathlib import Path
from re import search
from subprocess import PIPE  # noqa: S404
from subprocess import CalledProcessError  # noqa: S404
from subprocess import check_output  # noqa: S404

from beartype import beartype

from utilities.pathlib import PathLike


@beartype
def get_branch_name(*, cwd: PathLike = Path.cwd()) -> str:
    """Get the current branch name."""

    root = get_repo_root(cwd=cwd)
    output = check_output(  # noqa: S603, S607
        ["git", "rev-parse", "--abbrev-ref", "HEAD"],
        stderr=PIPE,
        cwd=root,
        text=True,
    )
    return output.strip("\n")


@beartype
def get_repo_name(*, cwd: PathLike = Path.cwd()) -> str:
    """Get the repo name."""

    root = get_repo_root(cwd=cwd)
    output = check_output(  # noqa: S603, S607
        ["git", "remote", "get-url", "origin"], stderr=PIPE, cwd=root, text=True
    )
    return Path(output.strip("\n")).stem


@beartype
def get_repo_root(*, cwd: PathLike = Path.cwd()) -> Path:
    """Get the repo root."""

    try:
        output = check_output(  # noqa: S603, S607
            ["git", "rev-parse", "--show-toplevel"],
            stderr=PIPE,
            cwd=cwd,
            text=True,
        )
    except CalledProcessError as error:
        if search("fatal: not a git repository", error.stderr):
            raise InvalidRepo(cwd) from None
        else:  # pragma: no cover
            raise
    else:
        return Path(output.strip("\n"))


class InvalidRepo(TypeError):
    ...



from collections.abc import Iterable
from collections.abc import Iterator
from functools import partial
from itertools import islice
from typing import TypeVar

from beartype import beartype


_T = TypeVar("_T")


@beartype
def take(n: int | None, iterable: Iterable[_T], /) -> list[_T]:
    """Return first n items of the iterable as a list."""

    return list(islice(iterable, n))


@beartype
def chunked(iterable: Iterable[_T], n: int | None, /) -> Iterator[Iterable[_T]]:
    """Return first n items of the iterable as a list."""

    return iter(partial(take, n, iter(iterable)), [])



from pytest import mark
from pytest import param

from utilities.beartype import IterableStrs
from utilities.itertools import chunked
from utilities.itertools import take


class TestChunked:
    @mark.parametrize(
        ["iterable", "n", "expected"],
        [
            param("ABCDEF", 3, [["A", "B", "C"], ["D", "E", "F"]]),
            param("ABCDE", 3, [["A", "B", "C"], ["D", "E"]]),
            param("ABCDE", None, [["A", "B", "C", "D", "E"]]),
        ],
    )
    def test_main(
        self, iterable: IterableStrs, n: int | None, expected: list[list[str]]
    ) -> None:
        assert list(chunked(iterable, n)) == expected


class TestTake:
    @mark.parametrize(
        ["n", "expected"],
        [
            param(0, []),
            param(1, [0]),
            param(2, [0, 1]),
            param(3, [0, 1, 2]),
            param(4, [0, 1, 2]),
        ],
    )
    def test_main(self, n: int, expected: list[int]) -> None:
        assert take(n, range(3)) == expected



from enum import auto
from enum import unique
from logging import basicConfig

from beartype import beartype

from utilities.enum import StrEnum


@beartype
def basic_config() -> None:
    """Do the basic config."""

    basicConfig(
        format="{asctime} | {name} | {levelname:8} | {message}",
        datefmt="%4Y-%m-%d %H:%M:%S",
        style="{",
        level="DEBUG",
    )


@unique
class LogLevel(StrEnum):
    """An enumeration of the logging levels."""

    DEBUG = auto()
    INFO = auto()
    WARNING = auto()
    ERROR = auto()
    CRITICAL = auto()



from collections.abc import Callable
from dataclasses import dataclass
from functools import wraps
from typing import Any
from typing import Generic
from typing import ParamSpec
from typing import TypeVar
from typing import cast

from beartype import beartype
from memory_profiler import memory_usage


_P = ParamSpec("_P")
_T = TypeVar("_T")


@beartype
@dataclass(frozen=True)
class Output(Generic[_T]):
    """A function output, and its memory usage."""

    value: _T
    memory: float


@beartype
def memory_profiled(func: Callable[_P, _T], /) -> Callable[_P, Output[_T]]:
    """Call a function, but also profile its maximum memory usage."""

    @wraps(func)
    def wrapped(*args: _P.args, **kwargs: _P.kwargs) -> Output[_T]:
        memory, value = memory_usage(
            cast(Any, (func, args, kwargs)), max_usage=True, retval=True
        )
        return Output(value=value, memory=memory)

    return wrapped



from typing import Any

from beartype import beartype


@beartype
def get_class_name(x: Any, /) -> str:
    """Get the name of a class."""

    return (x if isinstance(x, type) else type(x)).__name__
